import numpy as np
import pytest
from skimage import measure

from autolens import exc, dimensions as dim
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.data.array import grids
from autolens.data.array.util import grid_util

from test.unit.mock.model import mock_cosmology


class TestLightProfiles(object):
    class TestIntensity:
        def test__no_light_profiles__intensities_returned_as_0s_of_shape_grid(
            self, grid_stack_7x7
        ):

            galaxy = g.Galaxy(redshift=0.5)

            intensities = galaxy.intensities_from_grid(grid=grid_stack_7x7.regular)

            assert (
                intensities == np.zeros(shape=grid_stack_7x7.regular.shape[0])
            ).all()

        def test__using_no_light_profiles__check_reshaping_decorator_of_returned_intensities(
            self, grid_stack_7x7
        ):
            galaxy = g.Galaxy(redshift=0.5)

            intensities = galaxy.intensities_from_grid(
                grid=grid_stack_7x7.regular, return_in_2d=True, return_binned=False
            )

            assert (intensities == np.zeros(shape=(7, 7))).all()

            intensities = galaxy.intensities_from_grid(
                grid=grid_stack_7x7.sub, return_in_2d=False, return_binned=False
            )

            assert (intensities == np.zeros(shape=grid_stack_7x7.sub.shape[0])).all()

            intensities = galaxy.intensities_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert (
                intensities == np.zeros(shape=grid_stack_7x7.regular.shape[0])
            ).all()

        def test__galaxies_with_x1_and_x2_light_profiles__intensities_is_same_individual_profiles(
            self, lp_0, gal_x1_lp, lp_1, gal_x2_lp
        ):

            lp_intensities = lp_0.intensities_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_lp_intensities = gal_x1_lp.intensities_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert lp_intensities == gal_lp_intensities

            lp_intensities = lp_0.intensities_from_grid(grid=np.array([[1.05, -0.55]]))
            lp_intensities += lp_1.intensities_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_intensities = gal_x2_lp.intensities_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert lp_intensities == gal_intensities

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_light_profile_values(
            self, grid_stack_7x7, gal_x2_lp
        ):

            lp_0_intensities = gal_x2_lp.light_profile_0.intensities_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            lp_1_intensities = gal_x2_lp.light_profile_1.intensities_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            lp_intensities = lp_0_intensities + lp_1_intensities

            lp_intensities_0 = (
                lp_intensities[0]
                + lp_intensities[1]
                + lp_intensities[2]
                + lp_intensities[3]
            ) / 4.0

            lp_intensities_1 = (
                lp_intensities[4]
                + lp_intensities[5]
                + lp_intensities[6]
                + lp_intensities[7]
            ) / 4.0

            gal_intensities = gal_x2_lp.intensities_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert gal_intensities[0] == lp_intensities_0
            assert gal_intensities[1] == lp_intensities_1

    class TestLuminosityWithin:
        def test__two_profile_galaxy__is_sum_of_individual_profiles(
            self, lp_0, lp_1, gal_x1_lp, gal_x2_lp
        ):

            radius = dim.Length(0.5, "arcsec")

            lp_luminosity = lp_0.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="eps"
            )
            gal_luminosity = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="eps"
            )

            assert lp_luminosity == gal_luminosity

            lp_luminosity = lp_0.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity="eps"
            )
            lp_luminosity += lp_1.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity="eps"
            )

            gal_luminosity = gal_x2_lp.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity="eps"
            )

            assert lp_luminosity == gal_luminosity

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(
            self, lp_0, gal_x1_lp
        ):

            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=0.5, kpc_per_arcsec=2.0
            )

            radius = dim.Length(0.5, "arcsec")

            lp_luminosity_arcsec = lp_0.luminosity_within_circle_in_units(radius=radius)
            gal_luminosity_arcsec = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius
            )

            assert lp_luminosity_arcsec == gal_luminosity_arcsec

            radius = dim.Length(0.5, "kpc")

            lp_luminosity_kpc = lp_0.luminosity_within_circle_in_units(
                radius=radius, redshift_profile=0.5, cosmology=cosmology
            )
            gal_luminosity_kpc = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, cosmology=cosmology
            )

            assert lp_luminosity_kpc == gal_luminosity_kpc

        def test__luminosity_unit_conversions__multiply_by_exposure_time(
            self, lp_0, gal_x1_lp
        ):

            radius = dim.Length(0.5, "arcsec")

            lp_luminosity_eps = lp_0.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity="eps", exposure_time=2.0
            )
            gal_luminosity_eps = gal_x1_lp.luminosity_within_ellipse_in_units(
                major_axis=radius, unit_luminosity="eps", exposure_time=2.0
            )

            assert lp_luminosity_eps == gal_luminosity_eps

            lp_luminosity_counts = lp_0.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="counts", exposure_time=2.0
            )

            gal_luminosity_counts = gal_x1_lp.luminosity_within_circle_in_units(
                radius=radius, unit_luminosity="counts", exposure_time=2.0
            )

            assert lp_luminosity_counts == gal_luminosity_counts

        def test__no_light_profile__returns_none(self):

            gal_no_lp = g.Galaxy(redshift=0.5, mass=mp.SphericalIsothermal())

            assert gal_no_lp.luminosity_within_circle_in_units(radius=1.0) == None
            assert gal_no_lp.luminosity_within_ellipse_in_units(major_axis=1.0) == None

    class TestSymmetricProfiles(object):
        def test_1d_symmetry(self):

            lp_0 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
            )

            lp_1 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 0),
            )

            gal_x2_lp = g.Galaxy(
                redshift=0.5, light_profile_0=lp_0, light_profile_1=lp_1
            )

            assert gal_x2_lp.intensities_from_grid(
                np.array([[0.0, 0.0]])
            ) == gal_x2_lp.intensities_from_grid(grid=np.array([[100.0, 0.0]]))
            assert gal_x2_lp.intensities_from_grid(
                np.array([[49.0, 0.0]])
            ) == gal_x2_lp.intensities_from_grid(grid=np.array([[51.0, 0.0]]))

        def test_2d_symmetry(self):

            lp_0 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
            )

            lp_1 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 0),
            )

            lp_2 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(0, 100),
            )

            lp_3 = lp.EllipticalSersic(
                axis_ratio=1.0,
                phi=0.0,
                intensity=1.0,
                effective_radius=0.6,
                sersic_index=4.0,
                centre=(100, 100),
            )

            gal_x4_lp = g.Galaxy(
                redshift=0.5,
                light_profile_0=lp_0,
                light_profile_1=lp_1,
                light_profile_3=lp_2,
                light_profile_4=lp_3,
            )

            assert gal_x4_lp.intensities_from_grid(
                grid=np.array([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_lp.intensities_from_grid(grid=np.array([[51.0, 0.0]])), 1e-5
            )

            assert gal_x4_lp.intensities_from_grid(
                grid=np.array([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.intensities_from_grid(grid=np.array([[0.0, 51.0]])), 1e-5
            )

            assert gal_x4_lp.intensities_from_grid(
                grid=np.array([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.intensities_from_grid(grid=np.array([[100.0, 51.0]])), 1e-5
            )

            assert gal_x4_lp.intensities_from_grid(
                grid=np.array([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_lp.intensities_from_grid(grid=np.array([[51.0, 51.0]])), 1e-5
            )


def critical_curve_via_magnification_from_galaxy_and_grid(galaxy, grid):

    magnification_2d = galaxy.magnification_from_grid(
        grid=grid, return_in_2d=True, return_binned=False
    )

    inverse_magnification_2d = 1 / magnification_2d

    critical_curves_indices = measure.find_contours(inverse_magnification_2d, 0)

    no_critical_curves = len(critical_curves_indices)
    contours = []
    critical_curves = []

    for jj in np.arange(no_critical_curves):

        contours.append(critical_curves_indices[jj])
        contour_x, contour_y = contours[jj].T
        pixel_coord = np.stack((contour_x, contour_y), axis=-1)

        critical_curve = grid.marching_squares_grid_pixels_to_grid_arcsec(
            grid_pixels=pixel_coord, shape=magnification_2d.shape
        )

        critical_curves.append(critical_curve)

    return critical_curves


def caustics_via_magnification_from_galaxy_and_grid(galaxy, grid):

    caustics = []

    critical_curves = critical_curve_via_magnification_from_galaxy_and_grid(
        galaxy=galaxy, grid=grid
    )

    for i in range(len(critical_curves)):

        critical_curve = critical_curves[i]

        deflections_1d = galaxy.deflections_from_grid(
            grid=critical_curve, return_in_2d=False, return_binned=False
        )

        caustic = critical_curve - deflections_1d

        caustics.append(caustic)

    return caustics


class TestMassProfiles(object):
    class TestConvergence:
        def test__no_mass_profiles__convergence_returned_as_0s_of_shape_grid(
            self, grid_stack_7x7
        ):

            galaxy = g.Galaxy(redshift=0.5)

            convergence = galaxy.convergence_from_grid(grid=grid_stack_7x7.regular)

            assert (
                convergence == np.zeros(shape=grid_stack_7x7.regular.shape[0])
            ).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_convergence(
            self, grid_stack_7x7
        ):

            galaxy = g.Galaxy(redshift=0.5)

            convergence = galaxy.convergence_from_grid(
                grid=grid_stack_7x7.regular, return_in_2d=True, return_binned=False
            )

            assert (convergence == np.zeros(shape=(7, 7))).all()

            convergence = galaxy.convergence_from_grid(
                grid=grid_stack_7x7.sub, return_in_2d=False, return_binned=False
            )

            assert (convergence == np.zeros(shape=grid_stack_7x7.sub.shape[0])).all()

            convergence = galaxy.convergence_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert (
                convergence == np.zeros(shape=grid_stack_7x7.regular.shape[0])
            ).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__convergence_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            mp_convergence = mp_0.convergence_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_mp_convergence = gal_x1_mp.convergence_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert mp_convergence == gal_mp_convergence

            mp_convergence = mp_0.convergence_from_grid(grid=np.array([[1.05, -0.55]]))
            mp_convergence += mp_1.convergence_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_convergence = gal_x2_mp.convergence_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert mp_convergence == gal_convergence

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, grid_stack_7x7, gal_x2_mp
        ):

            mp_0_convergence = gal_x2_mp.mass_profile_0.convergence_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_1_convergence = gal_x2_mp.mass_profile_1.convergence_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_convergence = mp_0_convergence + mp_1_convergence

            mp_convergence_0 = (
                mp_convergence[0]
                + mp_convergence[1]
                + mp_convergence[2]
                + mp_convergence[3]
            ) / 4.0

            mp_convergence_1 = (
                mp_convergence[4]
                + mp_convergence[5]
                + mp_convergence[6]
                + mp_convergence[7]
            ) / 4.0

            gal_convergence = gal_x2_mp.convergence_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert gal_convergence[0] == mp_convergence_0
            assert gal_convergence[1] == mp_convergence_1

    class TestPotential:
        def test__no_mass_profiles__potential_returned_as_0s_of_shape_grid(
            self, grid_stack_7x7
        ):
            galaxy = g.Galaxy(redshift=0.5)

            potential = galaxy.potential_from_grid(grid=grid_stack_7x7.regular)

            assert (potential == np.zeros(shape=grid_stack_7x7.regular.shape[0])).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_potential(
            self, grid_stack_7x7
        ):
            galaxy = g.Galaxy(redshift=0.5)

            potential = galaxy.potential_from_grid(
                grid=grid_stack_7x7.regular, return_in_2d=True, return_binned=False
            )

            assert (potential == np.zeros(shape=(7, 7))).all()

            potential = galaxy.potential_from_grid(
                grid=grid_stack_7x7.sub, return_in_2d=False, return_binned=False
            )

            assert (potential == np.zeros(shape=grid_stack_7x7.sub.shape[0])).all()

            potential = galaxy.potential_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert (potential == np.zeros(shape=grid_stack_7x7.regular.shape[0])).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__potential_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            mp_potential = mp_0.potential_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_mp_potential = gal_x1_mp.potential_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert mp_potential == gal_mp_potential

            mp_potential = mp_0.potential_from_grid(grid=np.array([[1.05, -0.55]]))
            mp_potential += mp_1.potential_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_potential = gal_x2_mp.potential_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert mp_potential == gal_potential

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, grid_stack_7x7, gal_x2_mp
        ):
            mp_0_potential = gal_x2_mp.mass_profile_0.potential_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_1_potential = gal_x2_mp.mass_profile_1.potential_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_potential = mp_0_potential + mp_1_potential

            mp_potential_0 = (
                mp_potential[0] + mp_potential[1] + mp_potential[2] + mp_potential[3]
            ) / 4.0

            mp_potential_1 = (
                mp_potential[4] + mp_potential[5] + mp_potential[6] + mp_potential[7]
            ) / 4.0

            gal_potential = gal_x2_mp.potential_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert gal_potential[0] == mp_potential_0
            assert gal_potential[1] == mp_potential_1

    class TestDeflectionAngles:
        def test__no_mass_profiles__deflections_returned_as_0s_of_shape_grid(
            self, grid_stack_7x7
        ):

            galaxy = g.Galaxy(redshift=0.5)

            deflections = galaxy.deflections_from_grid(grid=grid_stack_7x7.regular)

            assert (
                deflections == np.zeros(shape=(grid_stack_7x7.regular.shape[0], 2))
            ).all()

        def test__using_no_mass_profiles__check_reshaping_decorator_of_returned_deflections(
            self, grid_stack_7x7
        ):

            galaxy = g.Galaxy(redshift=0.5)

            deflections = galaxy.deflections_from_grid(
                grid=grid_stack_7x7.regular, return_in_2d=True, return_binned=False
            )

            assert (deflections == np.zeros(shape=(7, 7, 2))).all()

            deflections = galaxy.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_in_2d=False, return_binned=False
            )

            assert (
                deflections == np.zeros(shape=(grid_stack_7x7.sub.shape[0], 2))
            ).all()

            deflections = galaxy.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert (
                deflections == np.zeros(shape=(grid_stack_7x7.regular.shape[0], 2))
            ).all()

        def test__galaxies_with_x1_and_x2_mass_profiles__deflections_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            mp_deflections = mp_0.deflections_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_mp_deflections = gal_x1_mp.deflections_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert (mp_deflections == gal_mp_deflections).all()

            mp_deflections = mp_0.deflections_from_grid(grid=np.array([[1.05, -0.55]]))
            mp_deflections += mp_1.deflections_from_grid(grid=np.array([[1.05, -0.55]]))

            gal_deflections = gal_x2_mp.deflections_from_grid(
                grid=np.array([[1.05, -0.55]])
            )

            assert (mp_deflections == gal_deflections).all()

        def test__sub_grid_in__grid_is_mapped_to_image_grid_by_wrapper_by_binning_sum_of_mass_profile_values(
            self, grid_stack_7x7, gal_x2_mp
        ):

            mp_0_deflections = gal_x2_mp.mass_profile_0.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_1_deflections = gal_x2_mp.mass_profile_1.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_binned=False
            )

            mp_deflections = mp_0_deflections + mp_1_deflections

            mp_deflections_y_0 = (
                mp_deflections[0, 0]
                + mp_deflections[1, 0]
                + mp_deflections[2, 0]
                + mp_deflections[3, 0]
            ) / 4.0

            mp_deflections_y_1 = (
                mp_deflections[4, 0]
                + mp_deflections[5, 0]
                + mp_deflections[6, 0]
                + mp_deflections[7, 0]
            ) / 4.0

            gal_deflections = gal_x2_mp.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert gal_deflections[0, 0] == mp_deflections_y_0
            assert gal_deflections[1, 0] == mp_deflections_y_1

            mp_deflections_x_0 = (
                mp_deflections[0, 1]
                + mp_deflections[1, 1]
                + mp_deflections[2, 1]
                + mp_deflections[3, 1]
            ) / 4.0

            mp_deflections_x_1 = (
                mp_deflections[4, 1]
                + mp_deflections[5, 1]
                + mp_deflections[6, 1]
                + mp_deflections[7, 1]
            ) / 4.0

            gal_deflections = gal_x2_mp.deflections_from_grid(
                grid=grid_stack_7x7.sub, return_binned=True
            )

            assert gal_deflections[0, 1] == mp_deflections_x_0
            assert gal_deflections[1, 1] == mp_deflections_x_1

    class TestMassWithin:
        def test__two_profile_galaxy__is_sum_of_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):

            radius = dim.Length(0.5, "arcsec")

            mp_mass = mp_0.mass_within_circle_in_units(
                radius=radius, unit_mass="angular"
            )

            gal_mass = gal_x1_mp.mass_within_circle_in_units(
                radius=radius, unit_mass="angular"
            )

            assert mp_mass == gal_mass

            mp_mass = mp_0.mass_within_ellipse_in_units(
                major_axis=radius, unit_mass="angular"
            )
            mp_mass += mp_1.mass_within_ellipse_in_units(
                major_axis=radius, unit_mass="angular"
            )

            gal_mass = gal_x2_mp.mass_within_ellipse_in_units(
                major_axis=radius, unit_mass="angular"
            )

            assert mp_mass == gal_mass

        def test__radius_unit_conversions__multiply_by_kpc_per_arcsec(
            self, mp_0, gal_x1_mp
        ):

            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=0.5, kpc_per_arcsec=2.0, critical_surface_density=1.0
            )

            radius = dim.Length(0.5, "arcsec")

            mp_mass_arcsec = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_profile=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_arcsec = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_arcsec == gal_mass_arcsec

            radius = dim.Length(0.5, "kpc")

            mp_mass_kpc = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_profile=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_kpc = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_kpc == gal_mass_kpc

        def test__mass_unit_conversions__same_as_individual_profile(
            self, mp_0, gal_x1_mp
        ):

            cosmology = mock_cosmology.MockCosmology(
                arcsec_per_kpc=1.0, kpc_per_arcsec=1.0, critical_surface_density=2.0
            )

            radius = dim.Length(0.5, "arcsec")

            mp_mass_angular = mp_0.mass_within_ellipse_in_units(
                major_axis=radius,
                unit_mass="angular",
                redshift_profile=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_angular = gal_x1_mp.mass_within_ellipse_in_units(
                major_axis=radius,
                unit_mass="angular",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_angular == gal_mass_angular

            mp_mass_sol = mp_0.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_profile=0.5,
                redshift_source=1.0,
                cosmology=cosmology,
            )

            gal_mass_sol = gal_x1_mp.mass_within_circle_in_units(
                radius=radius,
                unit_mass="solMass",
                redshift_source=1.0,
                cosmology=cosmology,
            )
            assert mp_mass_sol == gal_mass_sol

        def test__no_mass_profile__returns_none(self):

            gal_no_mp = g.Galaxy(redshift=0.5, light=lp.SphericalSersic())

            assert (
                gal_no_mp.mass_within_circle_in_units(
                    radius=1.0, critical_surface_density=1.0
                )
                == None
            )
            assert (
                gal_no_mp.mass_within_ellipse_in_units(
                    major_axis=1.0, critical_surface_density=1.0
                )
                == None
            )

    class TestSymmetricProfiles:
        def test_1d_symmetry(self):

            mp_0 = mp.EllipticalIsothermal(
                axis_ratio=0.5, phi=45.0, einstein_radius=1.0
            )

            mp_1 = mp.EllipticalIsothermal(
                centre=(100, 0), axis_ratio=0.5, phi=45.0, einstein_radius=1.0
            )

            gal_x4_mp = g.Galaxy(redshift=0.5, mass_profile_0=mp_0, mass_profile_1=mp_1)

            assert gal_x4_mp.convergence_from_grid(
                np.array([[1.0, 0.0]])
            ) == gal_x4_mp.convergence_from_grid(grid=np.array([[99.0, 0.0]]))

            assert gal_x4_mp.convergence_from_grid(
                np.array([[49.0, 0.0]])
            ) == gal_x4_mp.convergence_from_grid(grid=np.array([[51.0, 0.0]]))

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[1.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[99.0, 0.0]])), 1e-6
            )

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[51.0, 0.0]])), 1e-6
            )

            assert gal_x4_mp.deflections_from_grid(
                grid=np.array([[1.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[99.0, 0.0]])), 1e-6
            )

            assert gal_x4_mp.deflections_from_grid(
                grid=np.array([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[51.0, 0.0]])), 1e-6
            )

        def test_2d_symmetry(self):

            mp_0 = mp.SphericalIsothermal(einstein_radius=1.0)

            mp_1 = mp.SphericalIsothermal(centre=(100, 0), einstein_radius=1.0)

            mp_2 = mp.SphericalIsothermal(centre=(0, 100), einstein_radius=1.0)

            mp_3 = mp.SphericalIsothermal(centre=(100, 100), einstein_radius=1.0)

            gal_x4_mp = g.Galaxy(
                redshift=0.5,
                mass_profile_0=mp_0,
                mass_profile_1=mp_1,
                mass_profile_2=mp_2,
                mass_profile_3=mp_3,
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=np.array([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(grid=np.array([[51.0, 0.0]])), 1e-5
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=np.array([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(grid=np.array([[0.0, 51.0]])), 1e-5
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=np.array([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(grid=np.array([[100.0, 51.0]])), 1e-5
            )

            assert gal_x4_mp.convergence_from_grid(
                grid=np.array([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.convergence_from_grid(grid=np.array([[51.0, 51.0]])), 1e-5
            )

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[49.0, 0.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[51.0, 0.0]])), 1e-5
            )

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[0.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[0.0, 51.0]])), 1e-5
            )

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[100.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[100.0, 51.0]])), 1e-5
            )

            assert gal_x4_mp.potential_from_grid(
                grid=np.array([[49.0, 49.0]])
            ) == pytest.approx(
                gal_x4_mp.potential_from_grid(grid=np.array([[51.0, 51.0]])), 1e-5
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(grid=np.array([[49.0, 0.0]]))[
                0, 0
            ] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[51.0, 0.0]]))[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(grid=np.array([[0.0, 49.0]]))[
                0, 0
            ] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[0.0, 51.0]]))[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(
                grid=np.array([[100.0, 49.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[100.0, 51.0]]))[0, 0],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=np.array([[49.0, 49.0]])
            )[0, 0] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[51.0, 51.0]]))[0, 0],
                1e-5,
            )

            assert 1.0 * gal_x4_mp.deflections_from_grid(grid=np.array([[49.0, 0.0]]))[
                0, 1
            ] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[51.0, 0.0]]))[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(grid=np.array([[0.0, 49.0]]))[
                0, 1
            ] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[0.0, 51.0]]))[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=np.array([[100.0, 49.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[100.0, 51.0]]))[0, 1],
                1e-5,
            )

            assert -1.0 * gal_x4_mp.deflections_from_grid(
                grid=np.array([[49.0, 49.0]])
            )[0, 1] == pytest.approx(
                gal_x4_mp.deflections_from_grid(grid=np.array([[51.0, 51.0]]))[0, 1],
                1e-5,
            )

    class TestEinsteinRadiiMass:
        def test__x2_sis_different_einstein_radii_and_mass__einstein_radii_and_mass_are_sum(
            self
        ):

            mp_0 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
            mp_1 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=0.5)

            gal_x2_mp = g.Galaxy(redshift=0.5, mass_0=mp_0, mass_1=mp_1)

            assert gal_x2_mp.einstein_radius_in_units(unit_length="arcsec") == 1.5
            assert gal_x2_mp.einstein_mass_in_units(unit_mass="angular") == np.pi * (
                1.0 + 0.5 ** 2.0
            )

        def test__includes_shear__does_not_impact_values(self):

            mp_0 = mp.SphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)
            shear = mp.ExternalShear()

            gal_shear = g.Galaxy(redshift=0.5, mass_0=mp_0, shear=shear)

            assert gal_shear.einstein_radius_in_units(unit_length="arcsec") == 1.0
            assert gal_shear.einstein_mass_in_units(unit_mass="angular") == np.pi

    class TestDeflectionAnglesviaPotential(object):
        def test__compare_galaxy_deflections_via_potential_and_calculation(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.05
            )

            deflections_via_calculation = galaxy.deflections_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            deflections_via_potential = galaxy.deflections_via_potential_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            mean_error = np.mean(
                deflections_via_potential - deflections_via_calculation
            )

            assert mean_error < 1e-4

        def test__compare_two_component_galaxy_deflections_via_potential_and_calculation(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.05
            )

            deflections_via_calculation = galaxy.deflections_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            deflections_via_potential = galaxy.deflections_via_potential_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            mean_error = np.mean(
                deflections_via_potential - deflections_via_calculation
            )

            assert mean_error < 1e-4

        def test__galaxies_with_x1_and_x2_mass_profiles__deflections_via_potential_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05, sub_grid_size=2
            )

            mp_deflections = mp_0.deflections_via_potential_from_grid(grid=grid)

            gal_mp_deflections = gal_x1_mp.deflections_via_potential_from_grid(
                grid=grid
            )

            mean_error = np.mean(mp_deflections - gal_mp_deflections)

            assert mean_error < 1e-4

            mp_deflections = mp_0.deflections_via_potential_from_grid(grid=grid)
            mp_deflections += mp_1.deflections_via_potential_from_grid(grid=grid)

            gal_deflections = gal_x2_mp.deflections_via_potential_from_grid(grid=grid)

            mean_error = np.mean(mp_deflections - gal_deflections)

            assert mean_error < 1e-4

    class TestConvergenceviaJacobian(object):
        def test__compare_galaxy_convergence_via_jacobian_and_calculation(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05
            )

            convergence_via_calculation = galaxy.convergence_from_grid(
                grid=grid, return_in_2d=True, return_binned=True
            )

            convergence_via_jacobian = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=True, return_binned=True
            )

            mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

            assert mean_error < 1e-1

        def test__compare_two_component_galaxy_convergence_via_jacobian_and_calculation(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05
            )

            convergence_via_calculation = galaxy.convergence_from_grid(
                grid=grid, return_in_2d=True, return_binned=True
            )

            convergence_via_jacobian = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=True, return_binned=True
            )

            mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

            assert mean_error < 1e-1

        def test__convergence_sub_grid_binning_two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05, sub_grid_size=2
            )

            convergence_binned_reg_grid = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=False, return_binned=True
            )

            convergence_sub_grid = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=False, return_binned=False
            )

            pixel_1_reg_grid = convergence_binned_reg_grid[0]
            pixel_1_from_av_sub_grid = (
                convergence_sub_grid[0]
                + convergence_sub_grid[1]
                + convergence_sub_grid[2]
                + convergence_sub_grid[3]
            ) / 4

            assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

            pixel_10000_reg_grid = convergence_binned_reg_grid[99]

            pixel_10000_from_av_sub_grid = (
                convergence_sub_grid[399]
                + convergence_sub_grid[398]
                + convergence_sub_grid[397]
                + convergence_sub_grid[396]
            ) / 4

            assert pixel_10000_reg_grid == pytest.approx(
                pixel_10000_from_av_sub_grid, 1e-4
            )

            convergence_via_calculation = galaxy.convergence_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            convergence_via_jacobian = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=False, return_binned=True
            )

            mean_error = np.mean(convergence_via_jacobian - convergence_via_calculation)

            assert convergence_via_jacobian.shape == (400,)
            assert mean_error < 1e-1

        def test__galaxies_with_x1_and_x2_mass_profiles__convergence_via_jacobian_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05, sub_grid_size=2
            )

            mp_convergence = mp_0.convergence_from_jacobian(grid=grid)

            gal_mp_convergence = gal_x1_mp.convergence_from_jacobian(grid=grid)

            mean_error = np.mean(mp_convergence - gal_mp_convergence)

            assert mean_error < 1e-4

            mp_convergence = mp_0.convergence_from_jacobian(grid=grid)
            mp_convergence += mp_1.convergence_from_jacobian(grid=grid)

            gal_convergence = gal_x2_mp.convergence_from_jacobian(grid=grid)

            mean_error = np.mean(mp_convergence - gal_convergence)

            assert mean_error < 1e-4

    class TestShearviaJacobian(object):
        def test__galaxies_with_x1_and_x2_mass_profiles__shear_via_jacobian_is_same_individual_profiles(
            self, mp_0, gal_x1_mp, mp_1, gal_x2_mp
        ):
            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.05, sub_grid_size=2
            )

            mp_shear = mp_0.shear_from_jacobian(grid=grid)

            gal_mp_shear = gal_x1_mp.shear_from_jacobian(grid=grid)

            mean_error = np.mean(mp_shear - gal_mp_shear)

            assert mean_error < 1e-4

            mp_shear = mp_0.shear_from_jacobian(grid=grid)
            mp_shear += mp_1.shear_from_jacobian(grid=grid)

            gal_shear = gal_x2_mp.shear_from_jacobian(grid=grid)

            mean_error = np.mean(mp_shear - gal_shear)

            assert mean_error < 1e-4

        def test_shear_sub_grid_binning_two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
            )

            shear_binned_reg_grid = galaxy.shear_from_jacobian(
                grid=grid, return_in_2d=False, return_binned=True
            )

            shear_sub_grid = galaxy.shear_from_jacobian(
                grid=grid, return_in_2d=False, return_binned=False
            )

            pixel_1_reg_grid = shear_binned_reg_grid[0]
            pixel_1_from_av_sub_grid = (
                shear_sub_grid[0]
                + shear_sub_grid[1]
                + shear_sub_grid[2]
                + shear_sub_grid[3]
            ) / 4

            assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

            pixel_10000_reg_grid = shear_binned_reg_grid[99]

            pixel_10000_from_av_sub_grid = (
                shear_sub_grid[399]
                + shear_sub_grid[398]
                + shear_sub_grid[397]
                + shear_sub_grid[396]
            ) / 4

            assert pixel_10000_reg_grid == pytest.approx(
                pixel_10000_from_av_sub_grid, 1e-4
            )

    class TestJacobian(object):
        def test__jacobian_components(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05
            )

            jacobian = galaxy.lensing_jacobian_from_grid(grid=grid, return_in_2d=False)

            A_12 = jacobian[0, 1]
            A_21 = jacobian[1, 0]

            mean_error = np.mean(A_12 - A_21)

            assert mean_error < 1e-4

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
            )

            jacobian = galaxy.lensing_jacobian_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            A_12 = jacobian[0, 1]
            A_21 = jacobian[1, 0]

            mean_error = np.mean(A_12 - A_21)

            assert mean_error < 1e-4

        def test__jacobian_components__two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05
            )

            jacobian = galaxy.lensing_jacobian_from_grid(grid=grid, return_in_2d=False)

            A_12 = jacobian[0, 1]
            A_21 = jacobian[1, 0]

            mean_error = np.mean(A_12 - A_21)

            assert mean_error < 1e-4

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
            )

            jacobian = galaxy.lensing_jacobian_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )

            A_12 = jacobian[0, 1]
            A_21 = jacobian[1, 0]

            mean_error = np.mean(A_12 - A_21)

            assert mean_error < 1e-4

        def test__jacobian_sub_grid_binning_two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
            )

            jacobian_binned_reg_grid = galaxy.lensing_jacobian_from_grid(
                grid=grid, return_in_2d=False, return_binned=True
            )
            a11_binned_reg_grid = jacobian_binned_reg_grid[0, 0]

            jacobian_sub_grid = galaxy.lensing_jacobian_from_grid(
                grid=grid, return_in_2d=False, return_binned=False
            )
            a11_sub_grid = jacobian_sub_grid[0, 0]

            pixel_1_reg_grid = a11_binned_reg_grid[0]
            pixel_1_from_av_sub_grid = (
                a11_sub_grid[0] + a11_sub_grid[1] + a11_sub_grid[2] + a11_sub_grid[3]
            ) / 4

            assert jacobian_binned_reg_grid.shape == (2, 2, 100)
            assert jacobian_sub_grid.shape == (2, 2, 400)
            assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

            pixel_10000_reg_grid = a11_binned_reg_grid[99]

            pixel_10000_from_av_sub_grid = (
                a11_sub_grid[399]
                + a11_sub_grid[398]
                + a11_sub_grid[397]
                + a11_sub_grid[396]
            ) / 4

            assert pixel_10000_reg_grid == pytest.approx(
                pixel_10000_from_av_sub_grid, 1e-4
            )

        def test_lambda_t_sub_grid_binning_two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.05, sub_grid_size=2
            )

            lambda_t_binned_reg_grid = galaxy.tangential_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=False, return_binned=True
            )

            lambda_t_sub_grid = galaxy.tangential_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=False, return_binned=False
            )

            pixel_1_reg_grid = lambda_t_binned_reg_grid[0]
            pixel_1_from_av_sub_grid = (
                lambda_t_sub_grid[0]
                + lambda_t_sub_grid[1]
                + lambda_t_sub_grid[2]
                + lambda_t_sub_grid[3]
            ) / 4

            assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

            pixel_10000_reg_grid = lambda_t_binned_reg_grid[99]

            pixel_10000_from_av_sub_grid = (
                lambda_t_sub_grid[399]
                + lambda_t_sub_grid[398]
                + lambda_t_sub_grid[397]
                + lambda_t_sub_grid[396]
            ) / 4

            assert pixel_10000_reg_grid == pytest.approx(
                pixel_10000_from_av_sub_grid, 1e-4
            )

        def test_lambda_r_sub_grid_binning_two_component_galaxy(self):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
            )

            lambda_r_binned_reg_grid = galaxy.radial_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=False, return_binned=True
            )

            lambda_r_sub_grid = galaxy.radial_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=False, return_binned=False
            )

            pixel_1_reg_grid = lambda_r_binned_reg_grid[0]
            pixel_1_from_av_sub_grid = (
                lambda_r_sub_grid[0]
                + lambda_r_sub_grid[1]
                + lambda_r_sub_grid[2]
                + lambda_r_sub_grid[3]
            ) / 4

            assert pixel_1_reg_grid == pytest.approx(pixel_1_from_av_sub_grid, 1e-4)

            pixel_10000_reg_grid = lambda_r_binned_reg_grid[99]

            pixel_10000_from_av_sub_grid = (
                lambda_r_sub_grid[399]
                + lambda_r_sub_grid[398]
                + lambda_r_sub_grid[397]
                + lambda_r_sub_grid[396]
            ) / 4

            assert pixel_10000_reg_grid == pytest.approx(
                pixel_10000_from_av_sub_grid, 1e-4
            )

    class TestMagnification(object):
        def test__compare_magnification_from_eigen_values_and_from_determinant__two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )
            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05
            )

            magnification_via_determinant = galaxy.magnification_from_grid(
                grid=grid, return_in_2d=True
            )

            tangential_eigen_value = galaxy.tangential_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=True
            )

            radal_eigen_value = galaxy.radial_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=True
            )

            magnification_via_eigen_values = 1 / (
                tangential_eigen_value * radal_eigen_value
            )

            mean_error = np.mean(
                magnification_via_determinant - magnification_via_eigen_values
            )

            assert mean_error < 1e-4

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
            )

            magnification_via_determinant = galaxy.magnification_from_grid(
                grid=grid, return_in_2d=True, return_binned=False
            )

            tangential_eigen_value = galaxy.tangential_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=True, return_binned=False
            )

            radal_eigen_value = galaxy.radial_eigen_value_from_shear_and_convergence(
                grid=grid, return_in_2d=True, return_binned=False
            )

            magnification_via_eigen_values = 1 / (
                tangential_eigen_value * radal_eigen_value
            )

            mean_error = np.mean(
                magnification_via_determinant - magnification_via_eigen_values
            )

            assert mean_error < 1e-4

        def test__compare_magnification_from_determinant_and_from_convergence_and_shear__two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )
            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05
            )

            magnification_via_determinant = galaxy.magnification_from_grid(
                grid=grid, return_in_2d=True
            )

            convergence = galaxy.convergence_from_jacobian(grid=grid, return_in_2d=True)

            shear = galaxy.shear_from_jacobian(grid=grid, return_in_2d=True)

            magnification_via_convergence_and_shear = 1 / (
                (1 - convergence) ** 2 - shear ** 2
            )

            mean_error = np.mean(
                magnification_via_determinant - magnification_via_convergence_and_shear
            )

            assert mean_error < 1e-4

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(100, 100), pixel_scale=0.05, sub_grid_size=2
            )

            magnification_via_determinant = galaxy.magnification_from_grid(
                grid=grid, return_in_2d=True, return_binned=False
            )

            convergence = galaxy.convergence_from_jacobian(
                grid=grid, return_in_2d=True, return_binned=False
            )

            shear = galaxy.shear_from_jacobian(
                grid=grid, return_in_2d=True, return_binned=False
            )

            magnification_via_convergence_and_shear = 1 / (
                (1 - convergence) ** 2 - shear ** 2
            )

            mean_error = np.mean(
                magnification_via_determinant - magnification_via_convergence_and_shear
            )

            assert mean_error < 1e-4

    class TestCriticalCurvesandCaustics(object):
        def test__compare_tangential_critical_curves_from_magnification_and_lamda_t__reg_grid_two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.25
            )

            critical_curve_tangential_from_magnification = critical_curve_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                0
            ]

            critical_curve_tangential_from_lambda_t = galaxy.tangential_critical_curve_from_grid(
                grid=grid
            )

            assert critical_curve_tangential_from_lambda_t == pytest.approx(
                critical_curve_tangential_from_magnification, 5e-1
            )

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.5, sub_grid_size=2
            )

            critical_curve_tangential_from_magnification = critical_curve_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                0
            ]

            critical_curve_tangential_from_lambda_t = galaxy.tangential_critical_curve_from_grid(
                grid=grid
            )

            assert critical_curve_tangential_from_lambda_t == pytest.approx(
                critical_curve_tangential_from_magnification, 5e-1
            )

        def test__compare_radial_critical_curves_from_magnification_and_lamda_t__reg_grid_two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.25
            )

            critical_curve_radial_from_magnification = critical_curve_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                1
            ]

            critical_curve_radial_from_lambda_t = galaxy.radial_critical_curve_from_grid(
                grid=grid
            )

            assert sum(critical_curve_radial_from_lambda_t) == pytest.approx(
                sum(critical_curve_radial_from_magnification), 5e-1
            )

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.5, sub_grid_size=2
            )

            critical_curve_radial_from_magnification = critical_curve_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                1
            ]

            critical_curve_radial_from_lambda_t = galaxy.radial_critical_curve_from_grid(
                grid=grid
            )

            assert sum(critical_curve_radial_from_lambda_t) == pytest.approx(
                sum(critical_curve_radial_from_magnification), 5e-1
            )

        def test__compare_tangential_caustic_from_magnification_and_lambda_t__two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(20, 20), pixel_scale=0.25
            )

            caustic_tangential_from_magnification = caustics_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                0
            ]

            caustic_tangential_from_lambda_t = galaxy.tangential_caustic_from_grid(
                grid=grid
            )

            assert caustic_tangential_from_lambda_t == pytest.approx(
                caustic_tangential_from_magnification, 5e-1
            )

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(10, 10), pixel_scale=0.5, sub_grid_size=2
            )

            caustic_tangential_from_magnification = caustics_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                0
            ]

            caustic_tangential_from_lambda_t = galaxy.tangential_caustic_from_grid(
                grid=grid
            )

            assert caustic_tangential_from_lambda_t == pytest.approx(
                caustic_tangential_from_magnification, 5e-1
            )

        def test__compare_radial_caustic_from_magnification_and_lambda_t__two_component_galaxy(
            self
        ):
            mass_profile_1 = mp.SphericalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.0
            )
            mass_profile_2 = mp.SphericalIsothermal(
                centre=(1.0, 1.0), einstein_radius=1.0
            )

            galaxy = g.Galaxy(mass_1=mass_profile_1, mass_2=mass_profile_2, redshift=1)

            grid = grids.Grid.from_shape_pixel_scale_and_sub_grid_size(
                shape=(120, 120), pixel_scale=0.25, sub_grid_size=2
            )

            caustic_radial_from_magnification = caustics_via_magnification_from_galaxy_and_grid(
                galaxy=galaxy, grid=grid
            )[
                1
            ]
            caustic_radial_from_lambda_t = galaxy.radial_caustic_from_grid(grid=grid)

            assert sum(caustic_radial_from_lambda_t) == pytest.approx(
                sum(caustic_radial_from_magnification), 5e-1
            )


class TestMassAndLightProfiles(object):
    def test_single_profile(self, lmp_0):

        gal_x1_lmp = g.Galaxy(redshift=0.5, profile=lmp_0)

        assert 1 == len(gal_x1_lmp.light_profiles)
        assert 1 == len(gal_x1_lmp.mass_profiles)

        assert gal_x1_lmp.mass_profiles[0] == lmp_0
        assert gal_x1_lmp.light_profiles[0] == lmp_0

    def test_multiple_profile(self, lmp_0, lp_0, mp_0):

        gal_multi_profiles = g.Galaxy(redshift=0.5, profile=lmp_0, light=lp_0, sie=mp_0)

        assert 2 == len(gal_multi_profiles.light_profiles)
        assert 2 == len(gal_multi_profiles.mass_profiles)


class TestSummarizeInUnits(object):
    def test__galaxy_with_two_light_and_mass_profiles(self, lp_0, lp_1, mp_0, mp_1):

        gal_summarize = g.Galaxy(
            redshift=0.5,
            light_profile_0=lp_0,
            light_profile_1=lp_1,
            mass_profile_0=mp_0,
            mass_profile_1=mp_1,
        )

        summary_text = gal_summarize.summarize_in_units(
            radii=[dim.Length(10.0), dim.Length(500.0)],
            whitespace=50,
            unit_length="arcsec",
            unit_luminosity="eps",
            unit_mass="angular",
        )

        i = 0

        assert summary_text[i] == "Galaxy\n"
        i += 1
        assert (
            summary_text[i] == "redshift                                          0.50"
        )
        i += 1
        assert summary_text[i] == "\nGALAXY LIGHT\n\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    1.8854e+02 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   1.9573e+02 eps"
        )
        i += 1
        assert summary_text[i] == "\nLIGHT PROFILES:\n\n"
        i += 1
        assert summary_text[i] == "Light Profile = SphericalSersic\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    6.2848e+01 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   6.5243e+01 eps"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "Light Profile = SphericalSersic\n"
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_10.00_arcsec                    1.2570e+02 eps"
        )
        i += 1
        assert (
            summary_text[i]
            == "luminosity_within_500.00_arcsec                   1.3049e+02 eps"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "\nGALAXY MASS\n\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   3.00 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     1.5708e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          9.4248e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         4.7124e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\nMASS PROFILES:\n\n"
        i += 1
        assert summary_text[i] == "Mass Profile = SphericalIsothermal\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   1.00 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     3.1416e+00 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          3.1416e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         1.5708e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1
        assert summary_text[i] == "Mass Profile = SphericalIsothermal\n"
        i += 1
        assert (
            summary_text[i]
            == "einstein_radius                                   2.00 arcsec"
        )
        i += 1
        assert (
            summary_text[i]
            == "einstein_mass                                     1.2566e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_10.00_arcsec                          6.2832e+01 angular"
        )
        i += 1
        assert (
            summary_text[i]
            == "mass_within_500.00_arcsec                         3.1416e+03 angular"
        )
        i += 1
        assert summary_text[i] == "\n"
        i += 1


class TestHyperGalaxy(object):
    class TestContributionMaps(object):
        def test__model_image_all_1s__factor_is_0__contributions_all_1s(self):

            hyper_image = np.ones((3,))

            hyp = g.HyperGalaxy(contribution_factor=0.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image
            )

            assert (contribution_map == np.ones((3,))).all()

        def test__different_values__factor_is_1__contributions_are_value_divided_by_factor_and_max(
            self
        ):

            hyper_image = np.array([0.5, 1.0, 1.5])

            hyp = g.HyperGalaxy(contribution_factor=1.0)
            contribution_map = hyp.contribution_map_from_hyper_images(
                hyper_model_image=hyper_image, hyper_galaxy_image=hyper_image
            )

            assert (
                contribution_map
                == np.array([(0.5 / 1.5) / (1.5 / 2.5), (1.0 / 2.0) / (1.5 / 2.5), 1.0])
            ).all()

    class TestHyperNoiseMap(object):
        def test__contribution_all_1s__noise_factor_2__noise_adds_double(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.ones((3, 1))

            hyper_galaxy = g.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=1.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([2.0, 4.0, 6.0])).all()

        def test__same_as_above_but_contributions_vary(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = g.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=1.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([0.0, 2.0, 6.0])).all()

        def test__same_as_above_but_change_noise_scale_terms(self):

            noise_map = np.array([1.0, 2.0, 3.0])
            contribution_map = np.array([[0.0, 0.5, 1.0]])

            hyper_galaxy = g.HyperGalaxy(
                contribution_factor=0.0, noise_factor=2.0, noise_power=2.0
            )

            hyper_noise_map = hyper_galaxy.hyper_noise_map_from_contribution_map(
                noise_map=noise_map, contribution_map=contribution_map
            )

            assert (hyper_noise_map == np.array([0.0, 2.0, 18.0])).all()


class TestBooleanProperties(object):
    def test_has_profile(self):

        assert g.Galaxy(redshift=0.5).has_profile is False
        assert (
            g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_profile is True
        )
        assert g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_profile is True

    def test_has_light_profile(self):
        assert g.Galaxy(redshift=0.5).has_light_profile is False
        assert (
            g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_light_profile
            is True
        )
        assert (
            g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_light_profile
            is False
        )

    def test_has_mass_profile(self):
        assert g.Galaxy(redshift=0.5).has_mass_profile is False
        assert (
            g.Galaxy(redshift=0.5, light_profile=lp.LightProfile()).has_mass_profile
            is False
        )
        assert (
            g.Galaxy(redshift=0.5, mass_profile=mp.MassProfile()).has_mass_profile
            is True
        )

    def test_has_redshift(self):

        assert g.Galaxy(redshift=0.1).has_redshift is True

    def test_has_regularization(self):
        assert g.Galaxy(redshift=0.5).has_regularization is False
        assert (
            g.Galaxy(
                redshift=0.5, pixelization=object(), regularization=object()
            ).has_regularization
            is True
        )

    def test_has_hyper_galaxy(self):
        assert g.Galaxy(redshift=0.5, hyper_galaxy=object()).has_hyper_galaxy is True

    def test__only_pixelization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            g.Galaxy(redshift=0.5, pixelization=object())

    def test__only_regularization_raises_error(self):
        with pytest.raises(exc.GalaxyException):
            g.Galaxy(redshift=0.5, regularization=object())



import autofit as af
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.profiles import light_profiles as lp
from test.integration.tests import runner

test_type = "lens_only"
test_name = "lens_x2_galaxies"
data_type = "lens_only_x2_galaxies"
data_resolution = "LSST"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class LensPlanex2GalPhase(phase_imaging.PhaseImaging):
        def pass_priors(self, results):

            self.galaxies.lens_0.light.centre_0 = -1.0
            self.galaxies.lens_0.light.centre_1 = -1.0
            self.galaxies.lens_1.light.centre_0 = 1.0
            self.galaxies.lens_1.light.centre_1 = 1.0

    def mask_function(image):
        return msk.Mask.circular(
            shape=image.shape, pixel_scale=image.pixel_scale, radius_arcsec=5.0
        )

    phase1 = LensPlanex2GalPhase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens_0=gm.GalaxyModel(light=lp.EllipticalSersic),
            lens_1=gm.GalaxyModel(light=lp.EllipticalSersic),
        ),
        mask_function=mask_function,
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(name, phase1)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from autofit import conf
from autofit.tools import path_util
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import scaled_array
from autolens.data.array import mask as msk
from autolens.data.plotters import ccd_plotters

import numpy as np
import os

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

pixel_scale = 0.00976562
data_type = "subhalo_challenge_lp"
data_level = "level_0"
data_name = "small_hi_sn_system_1"

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_level, data_name]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
    resized_ccd_shape=(400, 400),
)

blurring_psf = abstract_data.PSF.from_gaussian(
    shape=(31, 31), pixel_scale=pixel_scale, sigma=50.0
)

blurred_image = blurring_psf.convolve(ccd_data.image)
blurred_image = scaled_array.ScaledSquarePixelArray(
    array=blurred_image, pixel_scale=pixel_scale
)
# array_plotters.plot_array(array=blurred_image)
absolute_blurred_signal_to_noise_map = np.abs(blurred_image) / ccd_data.noise_map
absolute_blurred_signal_to_noise_map = scaled_array.ScaledSquarePixelArray(
    array=absolute_blurred_signal_to_noise_map, pixel_scale=pixel_scale
)
# array_plotters.plot_array(array=absolute_blurred_signal_to_noise_map)

mask = np.where(absolute_blurred_signal_to_noise_map > 1.0, False, True)
mask = msk.Mask(array=mask, pixel_scale=pixel_scale)

ccd_plotters.plot_image(ccd_data=ccd_data, mask=mask)



from autolens.pipeline import phase_tagging


class TestPhaseTag:
    def test__mixture_of_values(self):

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_grid_size=2,
            signal_to_noise_limit=2,
            bin_up_factor=None,
            image_psf_shape=None,
            inversion_psf_shape=None,
            inner_mask_radii=0.3,
            positions_threshold=2.0,
            interp_pixel_scale=None,
            cluster_pixel_scale=None,
        )

        assert phase_tag == "phase_tag__sub_2__snr_2__pos_2.00__inner_mask_0.30"

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_grid_size=1,
            signal_to_noise_limit=None,
            bin_up_factor=3,
            image_psf_shape=(2, 2),
            inversion_psf_shape=(3, 3),
            inner_mask_radii=None,
            positions_threshold=None,
            interp_pixel_scale=0.2,
            cluster_pixel_scale=0.3,
        )

        assert (
            phase_tag
            == "phase_tag__sub_1__bin_3__image_psf_2x2__inv_psf_3x3__interp_0.200__cluster_0.300"
        )


class TestPhaseTaggers:
    def test__positions_threshold_tagger(self):

        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=None
        )
        assert tag == ""
        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=1.0
        )
        assert tag == "__pos_1.00"
        tag = phase_tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=2.56
        )
        assert tag == "__pos_2.56"

    def test__inner_circular_mask_radii_tagger(self):

        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=None
        )
        assert tag == ""
        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=0.2
        )
        print(tag)
        assert tag == "__inner_mask_0.20"
        tag = phase_tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=3
        )
        assert tag == "__inner_mask_3.00"

    def test__sub_grid_size_tagger(self):

        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=1)
        assert tag == "__sub_1"
        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=2)
        assert tag == "__sub_2"
        tag = phase_tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=4)
        assert tag == "__sub_4"

    def test__signal_to_noise_limit_tagger(self):

        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=None
        )
        assert tag == ""
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=1
        )
        assert tag == "__snr_1"
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=2
        )
        assert tag == "__snr_2"
        tag = phase_tagging.signal_to_noise_limit_tag_from_signal_to_noise_limit(
            signal_to_noise_limit=3
        )
        assert tag == "__snr_3"

    def test__bin_up_factor_tagger(self):

        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=None)
        assert tag == ""
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=1)
        assert tag == ""
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=2)
        assert tag == "__bin_2"
        tag = phase_tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=3)
        assert tag == "__bin_3"

    def test__image_psf_shape_tagger(self):

        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=None
        )
        assert tag == ""
        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=(2, 2)
        )
        assert tag == "__image_psf_2x2"
        tag = phase_tagging.image_psf_shape_tag_from_image_psf_shape(
            image_psf_shape=(3, 4)
        )
        assert tag == "__image_psf_3x4"

    def test__inversion_psf_shape_tagger(self):

        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=None
        )
        assert tag == ""
        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(2, 2)
        )
        assert tag == "__inv_psf_2x2"
        tag = phase_tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(3, 4)
        )
        assert tag == "__inv_psf_3x4"

    def test__interp_pixel_scale_tagger(self):

        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=None
        )
        assert tag == ""
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.5
        )
        assert tag == "__interp_0.500"
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.25
        )
        assert tag == "__interp_0.250"
        tag = phase_tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.234
        )
        assert tag == "__interp_0.234"

    def test__cluster_pixel_scale_tagger(self):

        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=None
        )
        assert tag == ""
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.5
        )
        assert tag == "__cluster_0.500"
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.25
        )
        assert tag == "__cluster_0.250"
        tag = phase_tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.234
        )
        assert tag == "__cluster_0.234"



from autofit.tools import path_util
from autolens.data.array import grids
from autolens.lens.plotters import ray_tracing_plotters
from autolens.data.plotters import ccd_plotters
from autolens.data.instrument import abstract_data
from autolens.data.instrument import ccd
from autolens.data.array import mask as msk
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.lens import lens_data as ld
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg
from autolens.lens.plotters import lens_fit_plotters

import os

workspace_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

# The pixel scale of the image to be simulated
pixel_scale = 0.1

# Simulate a simple Gaussian PSF for the image.
psf = abstract_data.PSF.from_gaussian(
    shape=(11, 11), sigma=0.1, pixel_scale=pixel_scale
)

# Setup the image-plane grid stack of the CCD array which will be used for generating the image-plane image of the
# simulated strong lens. The sub-grid size of 20x20 ensures we fully resolve the central regions of the lens and source
# galaxy light.
image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
    shape=(100, 100), pixel_scale=pixel_scale, sub_grid_size=16
)

# Setup the lens galaxy's light (elliptical Sersic), mass (SIE+Shear) and source galaxy light (elliptical Sersic) for
# this simulated lens.
lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
    ),
    shear=mp.ExternalShear(magnitude=0.05, phi=90.0),
)

source_galaxy = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=60.0,
        intensity=0.3,
        effective_radius=1.0,
        sersic_index=2.5,
    ),
)


# Use these galaxies to setup a tracer, which will generate the image-plane image for the simulated CCD instrument.
tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
    galaxies=[lens_galaxy, source_galaxy], image_plane_grid_stack=image_plane_grid_stack
)

# Lets look at the tracer's image-plane image - this is the image we'll be simulating.
# ray_tracing_plotters.plot_image_plane_image(tracer=tracer)

# Simulate the CCD instrument, remembering that we use a special image-plane image which ensures edge-effects don't
# degrade our modeling of the telescope optics (e.g. the PSF convolution).
ccd_data = ccd.SimulatedCCDData.from_image_and_exposure_arrays(
    image=tracer.padded_profile_image_plane_image_2d_from_psf_shape,
    pixel_scale=pixel_scale,
    exposure_time=300.0,
    psf=psf,
    background_sky_level=0.1,
    add_noise=True,
)

# Lets plot the simulated CCD instrument before we output it to files.
# ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

mask = msk.Mask.circular(
    shape=ccd_data.shape, pixel_scale=pixel_scale, radius_arcsec=3.0
)

lens_data = ld.LensData(ccd_data=ccd_data, mask=mask)

adaptive = pix.VoronoiMagnification(shape=(40, 40))

# Now lets plot our rectangular mapper with the image.
# mapper_plotters.plot_image_and_mapper(ccd_data=ccd_data, mapper=mapper, mask=mask, should_plot_grid=True)

# Okay, so lets think about the rectangular pixelization. Is it the optimal way to pixelize our source plane? Are there
# features in the source-plane that arn't ideal? How do you think we could do a better job?

# Well, given we're doing a whole tutorial on using a different pixelization to the rectangular grid, you've probably
# guessed that it isn't optimal. Infact, its pretty rubbish, and not a pixelization we'll actually want to model
# many lenses with!

# So what is wrong with the grid? Well, first, lets think about the source reconstruction.
# inversion_plotters.plot_reconstructed_pixelization(inversion=inversion, should_plot_centres=True)

source_galaxy = g.Galaxy(
    redshift=1.0, pixelization=adaptive, regularization=reg.Constant(coefficient=1.0)
)
tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
    galaxies=[lens_galaxy, source_galaxy], image_plane_grid_stack=lens_data.grid_stack
)
fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

print(fit.figure_of_merit)

# lens_fit_plotters.plot_fit_subplot(fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True)

source_galaxy = g.Galaxy(
    redshift=1.0,
    pixelization=adaptive,
    regularization=reg.AdaptiveBrightness(
        inner_coefficient=1.0, outer_coefficient=1.0, signal_scale=1.0
    ),
    hyper_model_image_1d=fit.model_image_1d,
    hyper_galaxy_image_1d=fit.model_image_1d,
    hyper_minimum_value=0.0,
)

tracer = ray_tracing.Tracer.from_galaxies_and_image_plane_grid_stack(
    galaxies=[lens_galaxy, source_galaxy], image_plane_grid_stack=lens_data.grid_stack
)
fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

print(fit.figure_of_merit)

lens_fit_plotters.plot_fit_subplot(
    fit=fit, should_plot_mask=True, extract_array_from_mask=True, zoom_around_mask=True
)



import numpy as np

from autolens.model.inversion import regularization
from autolens.model.inversion.util import regularization_util as reg_util


class TestRegularizationConstant:
    def test__regularization_matrix__compare_to_regularization_util(self):

        pixel_neighbors = np.array(
            [
                [1, 3, 7, 2],
                [4, 2, 0, -1],
                [1, 5, 3, -1],
                [4, 6, 0, -1],
                [7, 1, 5, 3],
                [4, 2, 8, -1],
                [7, 3, 0, -1],
                [4, 8, 6, -1],
                [7, 5, -1, -1],
            ]
        )

        pixel_neighbors_size = np.array([4, 3, 3, 3, 4, 3, 3, 3, 2])

        reg = regularization.Constant(coefficient=1.0)
        regularization_matrix = reg.regularization_matrix_from_pixel_neighbors(
            pixel_neighbors, pixel_neighbors_size
        )

        regularization_matrix_util = reg_util.constant_regularization_matrix_from_pixel_neighbors(
            coefficient=1.0,
            pixel_neighbors=pixel_neighbors,
            pixel_neighbors_size=pixel_neighbors_size,
        )

        assert (regularization_matrix == regularization_matrix_util).all()


class TestRegularizationWeighted:
    def test__pixel_signals__compare_to_regularization_util(self):

        reg = regularization.AdaptiveBrightness(signal_scale=2.0)

        sub_to_pix = np.array([0, 1, 2, 3, 3, 4, 5])
        sub_to_regular = np.array([0, 1, 2, 3, 3, 4, 5])
        galaxy_image = np.array([2.0, 1.0, 1.0, 4.0, 5.0, 6.0, 1.0])

        pixel_signals = reg.pixel_signals_from_images(
            pixels=6,
            sub_to_pix=sub_to_pix,
            sub_to_regular=sub_to_regular,
            hyper_image=galaxy_image,
        )

        pixel_signals_util = reg_util.adaptive_pixel_signals_from_images(
            pixels=6,
            signal_scale=2.0,
            sub_to_pix=sub_to_pix,
            sub_to_regular=sub_to_regular,
            hyper_image=galaxy_image,
        )

        assert (pixel_signals == pixel_signals_util).all()

    def test__weights__compare_to_regularization_util(self):

        reg = regularization.AdaptiveBrightness(
            inner_coefficient=10.0, outer_coefficient=15.0
        )

        pixel_signals = np.array([0.21, 0.586, 0.45])

        weights = reg.regularization_weights_from_pixel_signals(pixel_signals)

        weights_util = reg_util.adaptive_regularization_weights_from_pixel_signals(
            inner_coefficient=10.0, outer_coefficient=15.0, pixel_signals=pixel_signals
        )

        assert (weights == weights_util).all()

    def test__regularization_matrix__compare_to_regularization_util(self):

        reg = regularization.AdaptiveBrightness()

        pixel_neighbors = np.array(
            [
                [1, 4, -1, -1],
                [2, 4, 0, -1],
                [3, 4, 5, 1],
                [5, 2, -1, -1],
                [5, 0, 1, 2],
                [2, 3, 4, -1],
            ]
        )

        pixel_neighbors_size = np.array([2, 3, 4, 2, 4, 3])
        regularization_weights = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])

        regularization_matrix = reg.regularization_matrix_from_regularization_weights_and_pixel_neighbors(
            regularization_weights=regularization_weights,
            pixel_neighbors=pixel_neighbors,
            pixel_neighbors_size=pixel_neighbors_size,
        )

        regularization_matrix_util = reg_util.weighted_regularization_matrix_from_pixel_neighbors(
            regularization_weights=regularization_weights,
            pixel_neighbors=pixel_neighbors,
            pixel_neighbors_size=pixel_neighbors_size,
        )

        assert (regularization_matrix == regularization_matrix_util).all()



import time

from autolens.model.profiles import light_profiles as lp
from autolens.data.array import mask as msk
from autolens.lens import lens_data as ld

from test.simulation import simulation_util

# Although we could test the intensities without using an image (e.g. by just making a grid), we have chosen to
# set this test up using an image and mask. This gives run-time numbers that can be easily related to an actual lens
# analysis

sub_grid_size = 4
radius_arcsec = 3.0

print("sub grid size = " + str(sub_grid_size))
print("circular mask radius = " + str(radius_arcsec) + "\n")

for data_resolution in ["LSST", "Euclid", "HST", "HST_Up", "AO"]:

    ccd_data = simulation_util.load_test_ccd_data(
        data_type="no_lens_source_smooth",
        data_resolution=data_resolution,
        psf_shape=(3, 3),
    )
    mask = msk.Mask.circular(
        shape=ccd_data.shape,
        pixel_scale=ccd_data.pixel_scale,
        radius_arcsec=radius_arcsec,
    )
    lens_data = ld.LensData(ccd_data=ccd_data, mask=mask, sub_grid_size=sub_grid_size)

    print("Deflection angle run times for image type " + data_resolution + "\n")
    print("Number of points = " + str(lens_data.grid_stack.sub.shape[0]) + "\n")

    ### EllipticalGaussian ###

    mass_profile = lp.EllipticalGaussian(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, sigma=1.0
    )

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("EllipticalGaussian time = {}".format(diff))

    ### SphericalGaussian ###

    mass_profile = lp.SphericalGaussian(centre=(0.0, 0.0), sigma=1.0)

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("SphericalGaussian time = {}".format(diff))

    ### EllipticalExponential ###

    profile = lp.EllipticalExponential(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, intensity=1.0, effective_radius=1.0
    )

    start = time.time()
    profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("EllipticalExponential time = {}".format(diff))

    ### SphericalExponential ###

    profile = lp.SphericalExponential(
        centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0
    )

    start = time.time()
    profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("SphericalExponential time = {}".format(diff))

    ### EllipticalDevVaucouleurs ###

    profile = lp.EllipticalDevVaucouleurs(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=45.0, intensity=1.0, effective_radius=1.0
    )

    start = time.time()
    profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("EllipticalDevVaucouleurs time = {}".format(diff))

    ### SphericalDevVaucouleurs ###

    profile = lp.SphericalDevVaucouleurs(
        centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0
    )

    start = time.time()
    profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("SphericalDevVaucouleurs time = {}".format(diff))

    ### EllipticalSersic ###

    mass_profile = lp.EllipticalSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
    )

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("EllipticalSersic time = {}".format(diff))

    ### SphericalSersic ###

    mass_profile = lp.SphericalSersic(
        centre=(0.0, 0.0), intensity=1.0, effective_radius=1.0, sersic_index=2.5
    )

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("SphericalSersic time = {}".format(diff))

    ### EllipticalCoreSersic ###

    mass_profile = lp.EllipticalCoreSersic(
        centre=(0.0, 0.0),
        axis_ratio=0.8,
        phi=45.0,
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        radius_break=0.01,
        intensity_break=0.05,
        gamma=0.25,
        alpha=3.0,
    )

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("EllipticalCoreSersic time = {}".format(diff))

    ### SphericalCoreSersic ###

    mass_profile = lp.SphericalCoreSersic(
        centre=(0.0, 0.0),
        intensity=1.0,
        effective_radius=1.0,
        sersic_index=2.5,
        radius_break=0.01,
        intensity_break=0.05,
        gamma=0.25,
        alpha=3.0,
    )

    start = time.time()
    mass_profile.intensities_from_grid(grid=lens_data.grid_stack.sub)
    diff = time.time() - start
    print("SphericalCoreSersic time = {}".format(diff))

    print()



import numpy as np

from autolens.data.array.grids import reshape_returned_array, reshape_returned_grid


class MockGalaxy(object):
    def __init__(self, value, shape=1):
        self.value = value
        self.shape = shape

    @reshape_returned_array
    def intensities_from_grid(self, grid, return_in_2d=True, return_binned=True):
        return np.full(shape=self.shape, fill_value=self.value)

    @reshape_returned_array
    def convergence_from_grid(self, grid, return_in_2d=True, return_binned=True):
        return np.full(shape=self.shape, fill_value=self.value)

    @reshape_returned_array
    def potential_from_grid(self, grid, return_in_2d=True, return_binned=True):
        return np.full(shape=self.shape, fill_value=self.value)

    @reshape_returned_grid
    def deflections_from_grid(self, grid, return_in_2d=True, return_binned=True):
        return np.full(shape=(self.shape, 2), fill_value=self.value)


class MockHyperGalaxy(object):
    def __init__(self, contribution_factor=0.0, noise_factor=0.0, noise_power=1.0):
        self.contribution_factor = contribution_factor
        self.noise_factor = noise_factor
        self.noise_power = noise_power

    def contributions_from_model_image_and_galaxy_image(
        self, model_image, galaxy_image, minimum_value
    ):
        contributions = galaxy_image / (model_image + self.contribution_factor)
        contributions = contributions / np.max(contributions)
        contributions[contributions < minimum_value] = 0.0
        return contributions

    def hyper_noise_from_contributions(self, noise_map, contributions):
        return self.noise_factor * (noise_map * contributions) ** self.noise_power



from autolens.plotters import array_plotters
from autolens import exc


def plot_image(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    grid=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_image",
):
    """Plot the image of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.ccd.datas.CCD
        The datas-datas, which includes the observed datas, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    image = fit.image(return_in_2d=True)

    array_plotters.plot_array(
        array=image,
        grid=grid,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_noise_map(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_noise_map",
):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.ccd.datas.CCD
        The datas-datas, which includes the observed datas, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    noise_map = fit.noise_map(return_in_2d=True)

    array_plotters.plot_array(
        array=noise_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_signal_to_noise_map(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Signal-to-Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_signal_to_noise_map",
):
    """Plot the noise-map of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    image : datas.ccd.datas.CCD
    The datas-datas, which includes the observed datas, signal_to_noise_map-map, PSF, signal-to-signal_to_noise_map-map, etc.
    plot_origin : True
    If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    signal_to_noise_map = fit.signal_to_noise_map(return_in_2d=True)

    array_plotters.plot_array(
        array=signal_to_noise_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_model_data(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_mass_profile_centres=True,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Model Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_model_image",
):
    """Plot the model image of a fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the model image is plotted.
    """

    centres = get_mass_profile_centes(
        plot_mass_profile_centres=plot_mass_profile_centres, fit=fit
    )

    model_image = fit.model_image(return_in_2d=True)

    array_plotters.plot_array(
        array=model_image,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        centres=centres,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_residual_map(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Residuals",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_residual_map",
):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the residual_map are plotted.
    """

    residual_map = fit.residual_map(return_in_2d=True)

    array_plotters.plot_array(
        array=residual_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_normalized_residual_map(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Normalized Residuals",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_normalized_residual_map",
):
    """Plot the residual-map of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, normalized_residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the normalized_residual_map are plotted.
    """

    normalized_residual_map = fit.normalized_residual_map(return_in_2d=True)

    array_plotters.plot_array(
        array=normalized_residual_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_chi_squared_map(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Chi-Squareds",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_chi_squared_map",
):
    """Plot the chi-squared map of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the chi-squareds are plotted.
    """

    chi_squared_map = fit.chi_squared_map(return_in_2d=True)

    array_plotters.plot_array(
        array=chi_squared_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_subtracted_image_of_plane(
    fit,
    plane_index,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Fit Model Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_subtracted_image_of_plane",
):
    """Plot the model image of a specific plane of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the model image is plotted.
    plane_indexes : int
        The plane from which the model image is generated.
    """

    output_filename += "_" + str(plane_index)

    if fit.tracer.total_planes > 1:

        other_planes_model_images_2d = [
            model_image_2d
            for i, model_image_2d in enumerate(fit.model_image_2d_of_planes)
            if i != plane_index
        ]

        subtracted_image = fit.image(return_in_2d=True) - sum(
            other_planes_model_images_2d
        )

    else:

        subtracted_image = fit.image(return_in_2d=True)

    array_plotters.plot_array(
        array=subtracted_image,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        grid=image_plane_pix_grid,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_model_image_of_plane(
    fit,
    plane_index,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    plot_mass_profile_centres=True,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Model Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_model_image_of_plane",
):
    """Plot the model image of a specific plane of a lens fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractFitter
        The fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    plane_indexes : [int]
        The plane from which the model image is generated.
    """

    output_filename += "_" + str(plane_index)

    centres = get_mass_profile_centes(
        plot_mass_profile_centres=plot_mass_profile_centres, fit=fit
    )

    array_plotters.plot_array(
        array=fit.model_image_2d_of_planes[plane_index],
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        centres=centres,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_contribution_maps(
    fit,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Contributions",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="fit_contribution_maps",
):
    """Plot the summed contribution maps of a hyper_galaxy-fit.

    Set *autolens.datas.array.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractLensHyperFit
        The hyper_galaxy-fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
    image_index : int
        The index of the datas in the datas-set of which the contribution_maps are plotted.
    """

    if len(fit.contribution_maps) > 1:
        contribution_map = sum(fit.contribution_maps)
    else:
        contribution_map = fit.contribution_maps[0]

    array_plotters.plot_array(
        array=contribution_map,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def get_image_plane_pix_grid(should_plot_image_plane_pix, fit):

    if hasattr(fit, "inversion"):
        if (
            should_plot_image_plane_pix
            and fit.inversion.mapper.is_image_plane_pixelization
        ):
            return fit.tracer.image_plane.grid_stack.pixelization
    else:
        return None


def get_mask(fit, should_plot_mask):
    """Get the masks of the fit if the masks should be plotted on the fit.

    Parameters
    -----------
    fit : datas.fitting.fitting.AbstractLensHyperFit
        The fit to the datas, which includes a lisrt of every model image, residual_map, chi-squareds, etc.
    should_plot_mask : bool
        If *True*, the masks is plotted on the fit's datas.
    """
    if should_plot_mask:
        return fit.mask(return_in_2d=True)
    else:
        return None


def get_mass_profile_centes(plot_mass_profile_centres, fit):

    if not hasattr(fit, "tracer"):
        return None

    if plot_mass_profile_centres:
        return fit.tracer.image_plane.centres_of_galaxy_mass_profiles
    else:
        return None
